#!/usr/bin/env python
#
# build_es_schema.py
#
# Utility to formalize the build process for the elasticsearch schema and
# produce a consistent set of artifacts which can be used to initialize a
# new elasticsearch instance that will be compatible with the deployed code
#
# This deprecates the existing create_new_es_index* style scripts

import os
import tempfile
import sys
import logging
import json
import xml.etree.ElementTree as ET
import yaml
import shutil
import argparse

logging.basicConfig(format='%(asctime)s [%(levelname)s] %(message)s', level=logging.DEBUG)
PROJECT_ROOT = os.path.abspath(os.path.join(os.path.dirname(__file__), os.path.pardir, os.path.pardir, os.path.pardir))
RESOURCES = os.path.join(PROJECT_ROOT, 'src', 'main', 'resources')
MAPPINGS_DIR = os.path.join(RESOURCES, 'mappings')
SETTINGS_DIR = os.path.join(RESOURCES, 'settings')
DAO_CONFIG = os.path.join(RESOURCES, 'elasticsearch-dao.xml')
UPGRADE_SCRIPTS_DIR = os.path.join(PROJECT_ROOT, 'es')
ES_SETTINGS = os.path.join(os.path.dirname(__file__), 'default_configuration.yml')
OUTPUT_FILE = tempfile.NamedTemporaryFile(delete=False)
DEFAULT_OUTPUT_DIRECTORY = os.path.join(PROJECT_ROOT, 'target', 'elasticsearch')


def load_es_settings(settings_path):
    """
    Loads an elasticsearch settings yaml file
    
    :param settings_path: file to read 
    :return: settings dictionary
    """
    with open(settings_path) as document:
        settings = yaml.safe_load(document.read())
    return settings


def load_json_file(directory, filename):
    """
    Loads a json file and returns its content
    :param directory: Directory from which the file will be loaded
    :param filename: Filename to load
    :return: parsed content of the file
    """
    path = os.path.join(directory, filename)
    try:
        data = json.load(open(path))
        logging.debug("Loaded {0}".format(path))
    except ValueError as err:
        logging.error("{0} contains invalid json".format(path))
        logging.exception(err)
        sys.exit(1)

    return data


def save_json_file(directory, filename, data):
    """
    Writes out a pretty-printed json payload to a given file
    :param directory: Directory to which the file will be write
    :param filename: Filename to write
    :param data: Data to write
    """
    path = os.path.join(directory, filename)
    with open(path, "w") as output:
        json.dump(data, output, indent=2, sort_keys=True)


def make_clean_output_directory(output_directory):
    """
    Cleans the output directory for the build and creates the expected directory structure
    :param output_directory: target directory for build output 
    """
    logging.info("Clearing output directory {0} (if it exists)".format(output_directory))
    try:
        shutil.rmtree(output_directory)
    except OSError as e:
        # errno 2 is "No such file or directory"... expected
        if e.errno != 2:
            raise e
    os.mkdir(output_directory)
    logging.info("Creating new mappings and settings output directories")
    os.mkdir(os.path.join(output_directory, 'mappings'))
    os.mkdir(os.path.join(output_directory, 'settings'))
    os.mkdir(os.path.join(output_directory, 'upgrades'))
    shutil.copytree(os.path.join(os.path.dirname(__file__), 'es'), os.path.join(output_directory, 'es'))
    shutil.copy2(os.path.join(os.path.dirname(__file__), 'requirements.txt'), output_directory)
    shutil.copy2(os.path.join(os.path.dirname(__file__), 'provision_elasticsearch.py'), output_directory)
    shutil.copy2(os.path.join(os.path.dirname(__file__), 'upgrade_elasticsearch_schema.py'), output_directory)
    shutil.copy2(os.path.join(os.path.dirname(__file__), 'readme.md'), output_directory)


def update_settings_for_dao_configuration(settings, dao_config):
    """
    Reads the elasticsearch dao configuration file to determine important information 
    about the beans that use the elasticsearch indexes.
    
    This data is used to enrich the passed in settings object with the following keys:
    * index_name
    * is_template
    :param settings: 
    :param dao_config: 
    :return: 
    """
    tree = ET.parse(dao_config)
    beans = tree.findall(".//{http://www.springframework.org/schema/beans}property[@name='partitionStrategy']/..")
    bean_map = dict([(x["dao_name"], x) for x in settings])
    for bean in beans:
        bean_id = bean.attrib["id"]
        if bean_id not in bean_map:
            logging.info("Skipped bean " + bean_id)
            continue

        partition_strategy = bean.find("*[@name='partitionStrategy']")
        if partition_strategy is None:
            logging.error("Bean {0} missing partition strategy".format(bean_id))
            sys.exit(1)

        if "ref" in partition_strategy.attrib:
            beanName = partition_strategy.attrib["ref"]
            if ":" in beanName:
                beanName = beanName.split(":")[1].strip("}")
            selector = ".//{{http://www.springframework.org/schema/beans}}bean[@id='{0}']".format(
                beanName)
            strategy_bean = tree.find(selector)
        else:
            strategy_bean = list(partition_strategy)[0]

        strategy = strategy_bean.attrib["class"]

        if "index_name" not in bean_map[bean_id] or not bean_map[bean_id]["index_name"]:
            index_prefix_property = strategy_bean.find("*[@name='indexPrefix']")
            if index_prefix_property is not None:
                index_name = index_prefix_property.attrib["value"].split(":")[1].strip("}")
                bean_map[bean_id]["index_name"] = index_name
            else:
                logging.error("Bean {0} missing indexPrefix... cannot determine name".format(bean_id))
                sys.exit(1)

        bean_map[bean_id]["is_template"] = strategy != "com.scargo.backendservices.dao.impl.es.partition.DefaultPartitionStrategy"


def generate_output(es_settings, mappings_dir, settings_dir, upgrade_script_directory, output_directory):
    """
    Produces the build artifacts
    * Stores the default configuration yaml file
    * Validates all mappings and settings files and writes them to the target directory
    
    :param es_settings: Compiled settings object
    :param upgrade_script_directory: Location of upgrade scripts
    :param output_directory: Location for publishing artifacts 
    """
    mappings_output = os.path.join(output_directory, 'mappings')
    settings_output = os.path.join(output_directory, 'settings')
    for index_settings in es_settings["schema"]:
        if "mappings_file" in index_settings:
            mappings = load_json_file(mappings_dir, index_settings["mappings_file"])
            save_json_file(mappings_output, index_settings["mappings_file"], mappings)

        if "settings_file" in index_settings:
            settings = load_json_file(settings_dir, index_settings["settings_file"])
        else:
            settings = load_json_file(settings_dir, "config_settings.json")
            index_settings["settings_file"] = "{0}_settings.json".format(index_settings["index_name"])
        save_json_file(settings_output, index_settings["settings_file"], settings)

    with open(os.path.join(output_directory, "default_configuration.yml"), "w") as default:
        # remove default_index_configuration as changes to it will have no effect
        # writing out the new yaml makes the inheritance concrete
        del es_settings["default_index_configuration"]
        default.write(yaml.dump(es_settings, default_flow_style=False))

    if os.path.isdir(upgrade_script_directory):
        for directory in sorted(os.listdir(upgrade_script_directory)):
            if directory.lower().startswith("upgrade_"):
                src = os.path.join(upgrade_script_directory, directory)
                dest = os.path.join(output_directory, "upgrades", directory)
                logging.info("Copying upgrade directory {0} to {1}".format(src, dest))
                shutil.copytree(src, dest)


def duplicate_owner_perms(stat):
    """
    Returns an octal permissions mode where the original permissions of the 
    owner are mirrored to group and world access.
      
    :param stat: Original mode (int) 
    :return: new mode (int)
    """
    owner = stat & 0o700
    owner = oct(owner)[-3]  # work from the right since python 2.x uses 0777 and python 3.x uses 0o777
    desired = int("0o" + owner * 3, 8)
    return desired


def fix_permissions_for_jenkins(directory):
    """
    Helper method to update the permissions of the build artifacts.
    
    While most environments (local, gitlab) are able to properly package the 
    artifacts built by this script, the jenkins environment cannot.  The group 
    and world permissions are not properly set when copying files from the 
    docker volume containing the source to the artifacts directory.  
    
    This leads to trouble when attempting to tar the resources as the jenkins 
    build user cannot read the resulting artifacts.  
    
    As a workaround, we explicitly set the group and world permissions to be 
    the same as they are for the owner to ensure that the build user will be 
    able to read these files and produce the final archive.
    
    :param directory: directory to fix
    """
    for root, dirs, files in os.walk(directory):
        owner_perms = os.stat(root).st_mode
        desired_mode = owner_perms | duplicate_owner_perms(owner_perms)
        logging.info("chmod {0} {1}".format(oct(desired_mode), root))
        os.chmod(root, desired_mode)

        for file in files:
            file = os.path.join(root, file)
            file_perms = os.stat(file).st_mode
            desired_mode = file_perms | duplicate_owner_perms(file_perms)
            logging.info("chmod {0} {1}".format(oct(desired_mode), file))
            os.chmod(file, desired_mode)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--mappings-dir", default=MAPPINGS_DIR, help="Override the default mappings directory")
    parser.add_argument("--settings-dir", default=SETTINGS_DIR, help="Override the default settings directory")
    parser.add_argument("--output-dir", default=DEFAULT_OUTPUT_DIRECTORY, help="Override the default settings directory")
    parser.add_argument("--es-settings", default=ES_SETTINGS, help="Override the default es settings file")
    parser.add_argument("--dao-config", default=DAO_CONFIG, help="Override the default es settings file")
    args = parser.parse_args()
    print(args)

    es_settings = load_es_settings(args.es_settings)
    update_settings_for_dao_configuration(es_settings["schema"], args.dao_config)

    make_clean_output_directory(args.output_dir)
    generate_output(es_settings, args.mappings_dir, args.settings_dir, UPGRADE_SCRIPTS_DIR, args.output_dir)

    fix_permissions_for_jenkins(args.output_dir)


if __name__ == "__main__":
    main()
