import json
import copy


class IndexReport(object):

    # Report states
    UP_TO_DATE_MAPPINGS = "up_to_date_mappings"
    UP_TO_DATE_SETTINGS = "up_to_date_settings"
    MISSING_MAPPINGS = "missing_mappings"
    MISSING_SETTINGS = "missing_settings"
    DIFFERENT_MAPPINGS = "different_mappings"
    DIFFERENT_SETTINGS = "different_settings"

    """
    Helper class to store the validation results of an index/template
    
    Makes the judgement calls for high/low severity risks and reports status accordingly
    """

    def __init__(self, parent=None, name_override=None, index=None, strict_mappings_validation=True,
                 strict_settings_validation=True, is_missing=False, actual_mappings=None, actual_settings=None,
                 is_template=None):
        """
        Constructor.  Allows copying name and validation settings for use in recursive validation
        
        :param parent: Copy constructor IndexReport instance 
        :param index: The index this report is about
        :param strict_mappings_validation: whether strict mappings validation is enabled
        :param strict_settings_validation: whether strict settings validation is enabled
        :param is_missing: create the index report and immediately mark it as missing 
        :param actual_mappings: mappings as reported by ES
        :param actual_settings: settings as reported by ES
        :param is_template: manual override of template/index setting to allow indices 
        derived from templates to be reported as instantiated indices for correction
        """
        if parent is not None:
            self._index = parent._index
            self.strict_mappings_validation = parent.strict_mappings_validation
            self.strict_settings_validation = parent.strict_settings_validation

            # only store the child if a new name is provided (differentiate from nesting via properties/fields)
            if name_override:
                parent.children.append(self)
        else:
            self._index = copy.deepcopy(index)
            self.strict_mappings_validation = strict_mappings_validation
            self.strict_settings_validation = strict_settings_validation

        self.name_override = name_override
        self.is_template_override = is_template
        self.is_missing = is_missing
        self.children = []

        self.missing_fields = []
        self.missing_settings = []
        self.different_fields = []
        self.different_settings = []
        self.extra_fields = []
        self.extra_settings = []

        self.actual_mappings = actual_mappings
        self.actual_settings = actual_settings

    def add_missing_field(self, field_name, expected):
        """
        Records a missing field from an index's mappings
        
        :param field_name: name of the field 
        :param expected: expected json conf (from source)
        """
        self.missing_fields.append((field_name, "high", expected))

    def add_extra_field(self, field_name, actual):
        """
        Record an unexpected extra field that is present in ES but not the source mappings
        
        :param field_name: name of the field 
        :param actual: actual json conf (from ES)
        """
        self.extra_fields.append((field_name, "low", actual))

    def add_different_field(self, field_name, expected, actual):
        """
        Records a field that has a value differing from what is configured in source
        
        :param field_name: name of the field 
        :param expected: expected json conf (from source)
        :param actual: actual json conf (from ES)
        """
        if field_name in list(self._index["mappings"].values())[0]["properties"]:
            expected = list(self._index["mappings"].values())[0]["properties"][field_name]
        risk = "low"
        if expected.get("type", None) != actual.get("type", None):
            risk = "high"
        elif "fields" in expected and expected["fields"] != actual.get("fields", None):
            risk = "high"
        elif "properties" in expected and expected["properties"] != actual.get("properties"):
            risk = "high"

        self.different_fields.append((field_name, risk, expected, actual))

    def add_missing_setting(self, setting_name, expected):
        """
        Records a setting that is missing from ES
        
        :param setting_name: name of the setting 
        :param expected: expected value (from source)
        """
        risk = "high"
        if setting_name == "refresh_interval":
            risk = "low"
        self.missing_settings.append((setting_name, risk, expected))

    def add_different_setting(self, setting_name, expected, actual):
        """
        Records a setting that has a value differing from what is configured in source
        
        :param setting_name: name of the setting 
        :param expected: expected value (from source)
        :param actual: actual value (from ES)
        """
        risk = "high"
        if setting_name == "routing.allocation.exclude.tag" and expected == "raw_events" and actual == "":
            risk = "low"
        self.different_settings.append((setting_name, risk, expected, actual))

    def add_extra_setting(self, setting_name, actual):
        """
        Records an unexpected extra setting that is present in ES but not the source settings
        
        :param setting_name: name of the setting 
        :param actual: actual value (from ES)
        """
        risk = "high"

        # These cannot feasibly be in the actual settings on disk
        if setting_name not in ["creation_date", "version", "uuid"]:
            self.extra_settings.append((setting_name, risk, actual))

    def get_mappings_state(self):
        """
        Calculates the final judgement of an index's mappings.
        
        :return: IndexReport.MISSING_MAPPINGS, IndexReport.DIFFERENT_MAPPINGS, MAPPINGS_UP_TO_DATE
        """
        strict = self.strict_mappings_validation
        if self.is_missing:
            return IndexReport.MISSING_MAPPINGS
        if not self.missing_fields and not self.extra_fields and not self.different_fields:
            return IndexReport.UP_TO_DATE_MAPPINGS
        if strict:
            return IndexReport.DIFFERENT_MAPPINGS
        else:
            if any(missing[1] == "high" for missing in self.missing_fields):
                return IndexReport.DIFFERENT_MAPPINGS
            if any(different[1] == "high" for different in self.different_fields):
                return IndexReport.DIFFERENT_MAPPINGS
            return IndexReport.UP_TO_DATE_MAPPINGS

    def get_settings_state(self):
        """
        Calculates the final judgement of an index's mappings.
        
        :return: IndexReport.MISSING_MAPPINGS, IndexReport.DIFFERENT_MAPPINGS, MAPPINGS_UP_TO_DATE
        """
        strict = self.strict_settings_validation
        if self.is_missing:
            return IndexReport.MISSING_SETTINGS
        if not self.missing_settings and not self.different_settings and not self.extra_settings:
            return IndexReport.UP_TO_DATE_SETTINGS
        if strict:
            return IndexReport.DIFFERENT_SETTINGS
        else:
            if any(missing[1] == "high" for missing in self.missing_settings):
                return IndexReport.DIFFERENT_SETTINGS
            if any(different[1] == "high" for different in self.different_settings):
                return IndexReport.DIFFERENT_SETTINGS
            return IndexReport.UP_TO_DATE_SETTINGS

    @staticmethod
    def _short_json(verbosity, obj):
        """
        Helper method to conditionally shrink short json blobs to one line for clearer output
        
        :param verbosity: level of verbosity.  verbosity greater than one always does pretty printing 
        :param obj: object to print
        :return: string representation of obj
        """
        if not isinstance(obj, dict):
            return json.dumps(obj)
        if (len(obj) <= 2 and "properties" not in obj and "fields" not in obj) or verbosity <= 1:
            return json.dumps(obj, sort_keys=True)
        return json.dumps(obj, sort_keys=True, indent=2)

    def get_mappings_differences(self, verbosity=0):
        """
        Enumerates the differences between the expected index/template mappings and what 
        is actually in ES.
        
        Meant for visual representations.
        
        :param verbosity: level of verbosity at which the differences will be reported. 
        :return: list differences (as strings)  
        """
        strict = self.strict_mappings_validation

        for field in self.different_fields:
            if strict or field[1] == "high":
                yield "differing field " + field[0]
                if verbosity:
                    yield "  expected: " + IndexReport._short_json(verbosity, field[2])
                    yield "  actual: " + IndexReport._short_json(verbosity, field[3])

        for field in self.missing_fields:
            if strict or field[1] == "high":
                yield "missing field " + field[0]
                if verbosity:
                    yield "  expected: " + IndexReport._short_json(verbosity, field[2])

        for field in self.extra_fields:
            if strict or field[1] == "high":
                yield "unexpected field " + field[0]
                if verbosity:
                    yield "  actual: " + IndexReport._short_json(verbosity, field[2])

        first = True
        for child in self.children:
            problems = list(child.get_mappings_differences(verbosity))
            if problems:
                if first:
                    if not self.missing_fields and not self.different_fields and not self.extra_fields:
                        yield "template up to date"

                    yield "differing indices from template"
                    first = False

                yield "  {0}:".format(child.index_name)
                for problem in problems:
                    yield "    " + problem

    def get_settings_differences(self, verbosity=0):
        """
        Enumerates the differences between the expected index/template settings and what 
        is actually in ES.
        
        Meant for visual representations.
        
        :param verbosity: level of verbosity at which the differences will be reported. 
        :return: list of strings 
        """
        strict = self.strict_settings_validation

        for field in self.different_settings:
            if strict or field[1] == "high":
                yield "differing setting " + field[0]
                if verbosity:
                    yield "  expected: " + IndexReport._short_json(verbosity, field[2])
                    yield "  actual: " + IndexReport._short_json(verbosity, field[3])

        for field in self.missing_settings:
            if strict or field[1] == "high":
                yield "missing setting " + field[0]
                if verbosity:
                    yield "  expected: " + IndexReport._short_json(verbosity, field[2])

        for field in self.extra_settings:
            if strict or field[1] == "high":
                yield "unexpected setting " + field[0]
                if verbosity:
                    yield "  actual: " + IndexReport._short_json(verbosity, field[2])

    @property
    def index_name(self):
        return self.name_override or self.index["index_name"]

    @property
    def is_template(self):
        """
        Whether this report is for a specific index or a template
        :return: True/False
        """
        if self.is_template_override is not None:
            return self.is_template_override
        return self.index["is_template"]

    @property
    def index(self):
        """
        Don't give up our actual index since the verification process may 
        mutate it during processing.
        """
        return copy.deepcopy(self._index)

    def __str__(self):
        children = ""
        for child in self.children:
            children += "\t" + str(child)

        if children:
            children = "\n" + children
        return "[{0}] Mappings: {1}.  Settings: {2}{3}".format(self.index_name,
                                                               self.get_mappings_state(),
                                                               self.get_settings_state(),
                                                               children)
